#!/usr/bin/env ruby
#
# CS 430
# Introduction to Ruby
#
# Provides an intro to many Ruby concepts; intended to be used with
# the interactive Ruby interpreter (irb).
#
#


# obligatory math

3+4
2/3         # integer division by default
2.0/3       # floating-point
9%5         # modular division (remainder)
2**8        # exponentiation (^ is a bitwise operator)
2**100      # arbitrary integers


# everything is an object (also reflection)

5
5.class                 # base type
Fixnum.class            # Fixnum has a class
Class.class
Fixnum.superclass
Integer.superclass      # et cetera
(2**100).class          # Fixnum vs. Bignum
nil.class               # EVERYTHING is an object!

(5.0).class             # parentheses aren't strictly needed here
5.methods               # reflection; operators are methods
5.succ                  # "successor"
5.hash                  # everything can be hashed
5.send(:succ)           # calling a method = message passing
5.to_s                  # string conversion
5.to_s.class
5.methods.class


# arrays (closer to Python lists; uses dynamic resizing)

[1,2,3]
[1,2,3].class
[].class                        # empty array
[1, 2.3, "a"]                   # no type restrictions
[1,2,3].length                  # "size" and "count" also work (although count is slower)
[1,2,3].reverse                 # no parentheses for no-arg method calls
[1,2,3][0]                      # indices are zero-based
[1,2,3][-1]                     # negative indices work like in Python
[1,2,3].include?(2)             # boolean methods typically end in ?
[1,2,3].include?(5)             # parens aren't strictly necessary here
[1,2,3,4,5,6,7,8].slice(1,3)    # (start, len)
[1,4,2,3].sort
5.methods.sort
5.methods.sort - Object.methods
[1,2,3,4,5,6].shuffle
[1,2,3,4].first(2)              # extract first n elements
[1,2,3,4].drop(2)               # remove first n elements
[1,1,2,3,3,3].uniq              # discard duplicates


# strings (and more arrays)

"asdf".class                    # normal: expands escape codes and allows embedding
'asdf'.class                    # strict: only expands \'
"asdf".reverse
"hello world".include?("hello")
"a b c d".tr(" ", "_")
[1,2,3].join("-")
"a b c".split
"a,b,c".split
"a,b,c".split(",")
"ab"*3                          # repetition
"ab"+3                          # explicit conversion required here
"ab"+3.to_s
[1,2,3]+3                       # no meaning for "+" with arrays
[1,2,3]*3                       # repetition works for arrays too


# standard output

print("hi")
puts("hi")                          # newline output
puts "hi"                           # parentheses are optional!
puts 3                              # duck typing; "puts" calls "to_s" if present
puts 2.0/3.0
printf("%.3f\n", 2.0/3)             # works like in C
puts "a\tb"
puts 'a\tb'
puts "the number is " + 3.to_s      # explicit conversion required here
puts "double 3 is #{3*3}"           # embedding
puts 'double 3 is #{3*3}'           # doesn't work with single tick
puts gets
puts "in: \"#{gets}\""
puts "in: \"#{gets.chomp}\""        # strip trailing newline
puts [1,2,3]                        # prints each item on its own line
puts [1,2,3].inspect                # prints representation (like Python's "repr")
p [1,2,3]                           # "p x" is a shortcut for "puts x.inspect"


# blocks and higher-order functions

5.times do puts "hi" end            # loops & output
5.times   { puts "hi" }             # cleaner blocks
1.upto(5) { puts "hi" }             # 101 ways to do everything
[1,2,3].each { |x| puts x }         # block parameters (also implicit to_s)
[1,2,3].each { |x| puts 2**x }
[1,2,3].map  { |x| x+2 }            # functional programming!
5.methods.each { |m| puts m }
5.methods.each { |m| puts m } ;nil          # "each" returns the object
5.methods.sort.each { |m| puts m } ;nil


# booleans

true
false
3<4
3==4
"hi"=="hi"
true && true
true || false
true  ? "yes" : "no"    # ternary operator
false ? "yes" : "no"


# variables and assignments

n = 5
n += 2
a,b,c = 1,2,3       # parallel assignment
s = "123"
s.reverse
s.reverse!          # methods that modify the object end with "!"
s = gets.chomp


# regular expressions

a="quick brown fox"
a =~ /quick/            # matches at index 0
a =~ /brown/            # matches at index 6
a =~ /dog/              # no match
a =~ /^quick/           # beginning of line
a =~ /^brown/
a =~ /quick$/           # end of line
a =~ /dog$/
a.gsub(/[aeiou]/, 'x')
a.slice(/brown/)
a.slice!(/brown/)
a.slice!(/^quick/)


# ranges (like in Python) + more functional programming

1..20
(1..20).each   { |x| puts x }               # ranges support many features of arrays
(1..20).count  { |x| x%2==0 }               # want to extract even numbers
(1..20).each   { |x| puts x if x%2==0 }     # "if" clause
(1..20).map    { |x| x%2==0 ? x : 0 }
(1..20).map    { |x| x%2==0 ? x : 0 }.uniq
(1..20).map    { |x| x%2==0 ? x : 0 }.uniq.drop(1)
(1..20).select { |x| x%2==0 }


# hashes (really more of a "map" or "dictionary"; uses hashing, though)

h = Hash.new
h[1] = "one"
h[3] = "three"
h
h[1]
h.has_key?(2)
h.each_key { |k|   puts "#{k} = #{h[k]}" }      # iterates over keys
h.each     { |k,v| puts "#{k} = #{v}"    }      # iterates over key-value pairs


# constructors

Array.new
Array.new(5)
[nil]*5
Array.new(5,"a")


# I/O and files

File.open("test.txt", "w") { |f| f.puts "hello\nworld" }
File.foreach("test.txt") { |line| puts line }


#
# == In-class Exercises ===
#

# Write a line of Ruby code that prints fifteen powers of two, one
# per line, starting with "2". How many do you have memorized?

1.upto(15) { |p| puts 2**p }


# Write a line of Ruby code that returns all integers under 100 that
# are divisible by both three and seven. How many are there?

(1..100).select { |x| x%3==0 && x%7==0 }


# Write a line of Ruby code that reads through /proc/cpuinfo and
# prints every line that starts with "processor". How many processors
# does your computer have?

File.foreach("/proc/cpuinfo") { |line| puts line if line =~ /^processor/ }

